# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2018-03-17 17:41
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_competition', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='competition',
            name='voting_open',
            field=models.BooleanField(default=True, verbose_name='True if users can vote on the competition'),
        ),
        migrations.AlterField(
            model_name='competitionentry',
            name='description',
            field=models.TextField(help_text='You can use some markdown here.'),
        ),
    ]



# coding=utf-8
"""Urls."""
from django.conf.urls import url

from . import views

urlpatterns = [
  url(r'^done/?$', views.VoteSaved.as_view(), name='done'),
  url(r'^confirm/?$', views.ConfirmVote.as_view(), name='confirm'),
  url(
    r'vote/(?P<pk>[\-\d]+)/?$', views.CompetitionVote.as_view(), name='vote'
  )
]



# coding=utf-8
"""Command that recalculates votes."""
from django.core.management import CommandParser
from django.core.management.base import BaseCommand

from django_competition import services


class Command(BaseCommand):
  """Command that recalculates votes."""
  help = 'Re-calculate votes on competitions (in case of mismatch)'

  def add_arguments(self, parser: CommandParser):
    parser.add_argument('--competition-id', type=int)

  def handle(self, *args, **options):

    services.CompetitionService.recalculate_votes(options.get('competition-id'))



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2018-03-07 12:44
from __future__ import unicode_literals

import django.core.validators
import django.db.models.deletion
from django.db import migrations, models

import django_competition.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Competition',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=128)),
                ('description', models.TextField()),
                ('votes_per_day', models.IntegerField(default=-1, help_text='If > 0 allows single e-mail to vote this many votes in each day.If <= 0 allows only single vote.')),
                ('entries_per_vote', models.PositiveIntegerField(default=3, help_text='How many (equally weighted) entries you may select in each vote. ', validators=[django.core.validators.MinValueValidator(1)])),
                ('allow_disposable_emails', models.BooleanField(default=True, help_text='If false we will try to block disposable e-mails.')),
                ('competition_encryption_key', models.BinaryField(default=django_competition.models.random_key, help_text='Encryption key used to sign vote requests.')),
            ],
        ),
        migrations.CreateModel(
            name='CompetitionEntry',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=256)),
                ('description', models.TextField()),
                ('votes', models.IntegerField(default=0)),
                ('competition', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='entries', to='django_competition.Competition')),
            ],
        ),
        migrations.CreateModel(
            name='Vote',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('entry', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='entries', to='django_competition.CompetitionEntry')),
            ],
        ),
        migrations.CreateModel(
            name='VoteGroup',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('vote_source', models.TextField()),
                ('date_voted', models.DateTimeField(auto_now_add=True)),
                ('used_nonce', models.BinaryField(db_index=True)),
                ('competition', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_competition.Competition')),
            ],
        ),
        migrations.AddField(
            model_name='vote',
            name='group',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='votes', to='django_competition.VoteGroup'),
        ),
    ]



# coding=utf-8
"""Tests utilities related to saving votes and updating entry.votes attribute."""

import pytest
from django.core.management import call_command

from django_competition import services, models


# pylint: disable=redefined-outer-name,invalid-name,unused-argument


@pytest.fixture()
def persisted_vote(
    vote,
    competition: models.Competition
) -> services.CompetitionVote:
  """Vote stored in database."""
  service = services.CompetitionService(competition)
  service.register_vote(vote)
  return vote


@pytest.mark.django_db
def test_saved_vote_updates_entries(
    persisted_vote: services.CompetitionVote
):
  """Test entry.votes are updated when entry is saved."""
  for entry in persisted_vote.entries:
    entry.refresh_from_db()
    assert entry.votes == 1


@pytest.mark.django_db
def test_saved_vote_delete_updates_entries(
    persisted_vote: services.CompetitionVote
):
  """Test entry.votes are updated when entry is deleted."""
  for vote_group in models.VoteGroup.objects.all():
    vote_group.delete()
  for entry in persisted_vote.entries:
    entry.refresh_from_db()
    assert entry.votes == 0


def reset_competition_votes(competition):
  """
  Helper that sets votes to zero (but doesn't remove vote instants so after
  vote recalculation we should recover votes.
  """
  for entry in competition.entries.all():
    entry.votes = 0
    entry.save()


@pytest.mark.django_db
def test_synchronize_votes_single_competition(
    competition,
    other_competition,
    with_competition_votes
):
  """Test vote sync on single competition (other competition is not synced)"""
  reset_competition_votes(competition)
  reset_competition_votes(other_competition)
  services.CompetitionService.recalculate_votes(competition)
  for entry in competition.entries.all():
    assert entry.votes == abs(entry.pk)
  for entry in other_competition.entries.all():
    assert entry.votes == 0


@pytest.mark.django_db
def test_synchronize_votes_all_competitions(
    competition,
    other_competition,
    with_competition_votes
):
  """Test vote sync on all competitions."""
  reset_competition_votes(competition)
  reset_competition_votes(other_competition)
  services.CompetitionService.recalculate_votes()
  for entry in competition.entries.all():
    assert entry.votes == abs(entry.pk)
  for entry in other_competition.entries.all():
    assert entry.votes == abs(entry.pk)


@pytest.mark.django_db
def test_update_votes_command_single_competition(
    competition,
    other_competition,
    with_competition_votes
):
  """Test vote sync on single competition (using manage command)."""
  reset_competition_votes(competition)
  reset_competition_votes(other_competition)
  call_command("update_votes", **{"competition-id": competition.pk})
  for entry in competition.entries.all():
    assert entry.votes == abs(entry.pk)
  for entry in other_competition.entries.all():
    assert entry.votes == 0


@pytest.mark.django_db
def test_update_votes_command_both_competitions(
    competition,
    other_competition,
    with_competition_votes
):
  """Test vote sync on all competitions (using manage command)."""
  reset_competition_votes(competition)
  reset_competition_votes(other_competition)
  call_command("update_votes")
  for entry in competition.entries.all():
    assert entry.votes == abs(entry.pk)
  for entry in other_competition.entries.all():
    assert entry.votes == abs(entry.pk)



# coding=utf-8
"""Shared fixtures."""

import pytest

from django_competition import models, services


# pylint: disable=redefined-outer-name


@pytest.fixture()
def competition():
  """
  Main competition used in tests.

  Uses negative primary keys, so other tests can assume *known* primary keys and
  we don't have to

  """
  competition = models.Competition.objects.create(
    id=-1,
    name="Test",
    description="Test",
    votes_per_day=1,
    entries_per_vote=3,
    allow_disposable_emails=False,
    competition_encryption_key=64 * b'\x01'
  )

  for entry_id in range(4):
    models.CompetitionEntry.objects.create(
      id=-(1 + entry_id),
      competition=competition,
      name="Entry {}".format(entry_id),
      description=""
    )

  return competition


@pytest.fixture()
def other_competition():
  """Second competition."""
  competition = models.Competition.objects.create(
    name="Test",
    description="Test",
    votes_per_day=1,
    entries_per_vote=3,
    allow_disposable_emails=False,
    competition_encryption_key=64 * b'\x01'
  )

  for entry_id in range(4):
    models.CompetitionEntry.objects.create(
      competition=competition,
      name="Entry {}".format(entry_id),
      description=""
    )
  return competition


@pytest.fixture()
def with_competition_votes(competition, other_competition):
  """Adds votes to competitions, each entry has abs(entry.pk) votes."""
  def __make_votes(competition):
    for entry in competition.entries.all():
      for _ in range(abs(entry.pk)):
        vote_group = models.VoteGroup.objects.create(
          competition=competition, vote_source="test@somewhere.org"
        )
        models.Vote.objects.create(
          group=vote_group, entry=entry
        )

  __make_votes(competition)
  __make_votes(other_competition)


@pytest.fixture()
def other_entry(other_competition):
  """Entry on other competition"""
  return other_competition.entries.first()


@pytest.fixture()
def service():
  """Vote service."""
  return services.VoteService()


@pytest.fixture()
def vote(competition):
  """Valid competition vote."""
  return services.CompetitionVote(
    competition,
    entries=list(competition.entries.all().order_by('pk'))[2:],
    nonce=b'1234',
    source_email='none@no-such-email-provider.pl'
  )


@pytest.fixture()
def serialized_vote():
  """Serialized version of ``vote``"""
  return b"competition=-1&nonce=1234" \
         b"&source_email=none%40no-such-email-provider.pl&" \
         b"entry=-1&entry=-2"


@pytest.fixture()
def vote_signature():
  """Valid signature for serialized_vote."""
  return (
    b'\x17=\xe6\xfb\xe9\xc8\xe9:\xcd\xf9\x88\xad\xfe@\x86\x10\xa6h\xd4\xdd'
    b'\xa4"0\xcb}\xc0\x9b\x85\xbc)\xb8\x95'
  )


@pytest.fixture()
def valid_competition_form_data():
  """Valid data for competition form."""
  return {
    "choice_0": -1,
    "your_email": "user@gmail.com",
    "store_e_mail_consent": "on",
    "terms_and_condition_consents": "on",
  }


INVALID_COMPETITION_FORM_DATA = [
  # No email
  [
    {'choice_0': '-1', "store_e_mail_consent": "on", "terms_and_condition_consents": "on"},
    {"your_email"}
  ],
  # Invalid e-mail format
  [
    {
      'choice_0': '-1', "your_email": "Invalid email",
      "store_e_mail_consent": "on", "terms_and_condition_consents": "on",
    },
    {"your_email"}
  ],
  # Blocked e-mail domain
  [
    {
      'choice_0': '-1', "your_email": "jacek@niepodam.pl",
      "store_e_mail_consent": "on", "terms_and_condition_consents": "on",
    },
    {"__all__"}
  ],
  # No choice
  [
    {
      "your_email": "someone@no-such-domain.pl", "store_e_mail_consent": "on",
      "terms_and_condition_consents": "on"
    },
    {"choice_0"}
  ],
  # No consent
  [
    {
      'choice_0': '-1', "your_email": "someone@no-such-domain.pl",
      "terms_and_condition_consents": "on",
    },
    {"store_e_mail_consent"}
  ],
  # No consent
  [
    {'choice_0': '-1', "your_email": "someone@no-such-domain.pl", "store_e_mail_consent": "on", },
    {"terms_and_condition_consents"}
  ],
  # Repeated choices
  [
    {
      'choice_0': '-1',
      'choice_1': '-1',
      "your_email": "jacek@niepodam.pl",
      "store_e_mail_consent": "on",
      "terms_and_condition_consents": "on"
    },
    {"__all__"}
  ]
]


@pytest.fixture(params=INVALID_COMPETITION_FORM_DATA)
def invalid_competition_form_data(request):
  """
  Returns tuple form_data, broken_fields,
  where form_data a dictionary of form field values and
  broken fields is a set of fields that have errors.
  """
  return request.param



# coding=utf-8

"""Forms."""

import base64
import os
import typing
from collections import OrderedDict

from django import forms
from django.core.exceptions import ValidationError
from django.forms import ModelChoiceField
from django.forms.forms import DeclarativeFieldsMetaclass
from django.utils.safestring import SafeString
from django.utils.translation import ugettext_lazy

from django_competition.services import SETTINGS_SERVICE
from . import models, services


class BaseVoteCompetitionForm(forms.Form):
  """Base form, that contains validations."""

  your_email = forms.EmailField(
    required=True,
    label=ugettext_lazy("Your e-mail"),
  )

  def clean(self):

    if self.errors:
      # If there are errors already don't validate further
      return

    selected_entries = set()

    for name in self.fields.keys():
      if not name.startswith('choice_'):
        continue
      entry = self.cleaned_data.get(name)
      if entry is None:
        continue
      if entry in selected_entries:
        raise ValidationError(
          "You have selected the same entry twice. Entry is: {}".format(entry.name)
        )
      selected_entries.add(entry)

    vote = self.make_vote()
    try:
      services.VoteService().validate_vote(vote)
    except services.UserInvalidVoteException as exception:
      raise ValidationError(exception.args[0]) from exception

  def get_vote_entries(self):
    """Returns entries this vote voted on."""
    entries = []
    for name in self.fields.keys():
      if name.startswith('choice_'):
        value = self.cleaned_data[name]
        if value is not None:
          entries.append(value)
    return entries

  def make_vote(self):
    """Create a CompetitionVote instance."""
    return services.CompetitionVote(
      competition=self.competition,
      source_email=self.cleaned_data['your_email'],
      entries=self.get_vote_entries(),
      nonce=base64.b64encode(os.urandom(16))
    )


class CompetitionFormWithConfirmations(BaseVoteCompetitionForm):

  """Base form with extra consents."""

  store_e_mail_consent = forms.BooleanField(
    label=ugettext_lazy(
      "I agree that this platform will store my e-mail address and use it to check any "
      "fraudulent activity in the voting process."
    ),
    required=True
  )

  terms_and_condition_consents = forms.BooleanField(
    required=True,
  )

  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self.fields['terms_and_condition_consents'].label = SafeString(ugettext_lazy(
      "I agree to terms and conditions <a href=\"{}\">located here</a>".format(
        SETTINGS_SERVICE.terms_url
      )
    ))


class CompetitionEntryChoiceField(ModelChoiceField):

  """Field for competition entry in voting view."""

  def label_from_instance(self, obj: models.Competition):
    return obj.name


def create_vote_form(
    competition: models.Competition,
    form_base=CompetitionFormWithConfirmations
) -> typing.Type[BaseVoteCompetitionForm]:
  """
  Creates competition form form a base form.

  Adds choice fields that allow end-users to actually vote.
  :param competition: Competition for which to prepare form.
  :param form_base: Base form.
  """

  form_fields = [
    ('competition', competition)
  ]

  votes = min(
    competition.entries.count(),
    competition.entries_per_vote
  )

  form_fields.extend((
    (
      "choice_{choice}".format(choice=choice),
      CompetitionEntryChoiceField(
        required=choice == 0,
        queryset=competition.entries.all(),
        label=ugettext_lazy("Your choice"))
    )
    for choice in range(votes)
  ))

  return DeclarativeFieldsMetaclass(
    "CompetitionForm", (form_base, ), OrderedDict(form_fields)
  )



"""Views module"""
import random
from urllib.parse import urlencode

from django.http import HttpResponseRedirect, HttpResponse
from django.urls import reverse
from django.utils.safestring import SafeString
from django.views.generic import FormView, TemplateView, DetailView

from django_competition.services import (
  VoteService,
  UserInvalidVoteException,
  InvalidVoteException,
  SETTINGS_SERVICE,
)
from . import services, models, forms


class ConfirmVote(TemplateView):
  """View that asks user to confirm vote."""

  template_name = "django_competition/vote_confirm.html"

  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self.service = services.VoteService()
    self.vote = None
    self.form_error = None

  def get_context_data(self, **kwargs):
    kwargs.update({
      "vote": self.vote,
      "form_error": self.form_error
    })
    return super().get_context_data(**kwargs)

  def dispatch(self, *args, **kwargs):  # pylint: disable=arguments-differ
    try:
      self.vote = self.service.query_args_to_vote(self.request.GET)
    except UserInvalidVoteException as exception:
      self.form_error = exception.args[0]
      return self.get(*args, **kwargs)
    except InvalidVoteException:
      return HttpResponse(status=400)
    return super().dispatch(*args, **kwargs)

  def post(self, request):  # pylint: disable=unused-argument
    """Post handler, register vote and redirects."""
    competition_service = services.CompetitionService(self.vote.competition)
    # Note we know vote is now valid.
    competition_service.register_vote(self.vote)
    return HttpResponseRedirect(
      redirect_to=SETTINGS_SERVICE.vote_registered_url
    )


class VoteSaved(TemplateView):
  """Landing page after user voted but before confirmed."""

  template_name = "django_competition/vote_performed.html"


class CompetitionVote(DetailView, FormView):

  """Vote page."""

  model = models.Competition
  context_object_name = "competition"
  template_name = "django_competition/vote.html"

  def get_form_class(self):
    return forms.create_vote_form(
      self.object,
      SETTINGS_SERVICE.base_form
    )

  def get_context_data(self, **kwargs):
    ctx = super().get_context_data(**kwargs)
    entries = list(self.object.entries.all())
    random.shuffle(entries)
    ctx.update({
      "entries": entries,
    })
    return ctx

  def send_email(self, vote):
    """Send confirmation e-mail."""

    SETTINGS_SERVICE.mail_service(
      vote=vote,
      confirm_url=self.format_confirm_url(vote)
    ).send_confirm_mail()

  def format_confirm_url(self, vote) -> SafeString:
    """Prepare confirmation url."""
    service = VoteService()
    params = service.vote_to_signed_query_args(vote)
    query_string = urlencode(params, doseq=False)
    path = reverse("competition:confirm")
    return SafeString(
      self.request.build_absolute_uri(path) + "?" + query_string
    )

  def form_valid(self, form):
    self.send_email(form.make_vote())
    return super().form_valid(form)

  def get_success_url(self):
    return SETTINGS_SERVICE.vote_confirm_success_url

  def dispatch(self, request, *args, **kwargs):
    self.object = self.get_object()  # pylint: disable=attribute-defined-outside-init
    return super().dispatch(request, *args, **kwargs)



"""Services module."""

# coding=utf-8
import abc
import base64
import binascii
import hashlib
import hmac
import importlib
import typing
from urllib.parse import urlencode, parse_qs

import bleach
from MailChecker import MailChecker
from django.conf import settings
from django.core.mail import send_mail
from django.db import transaction
from django.template.loader import get_template
from django.utils.encoding import force_bytes
from django.utils.translation import ugettext_lazy

from . import models

DEFAULT_SETTINGS = {
  "MAIL_SERVICE": "django_competition.services.SimpleMailService",
  "COMPETITION_BASE_FORM": "django_competition.forms.CompetitionFormWithConfirmations",
  "BLEACH_CLEANER": "django_competition.services.BLEACH_CLEANER"
}


CompetitionVote = typing.NamedTuple(
  "CompetitionVote",
  (
    ("competition", 'models.Competition'),
    ("entries", typing.List['models.CompetitionEntry']),
    ("source_email", str),
    ("nonce", bytes),
  )
)

# pylint: disable=no-self-use


class SettingsService(object):

  """Service that encapsulates settings."""

  def __init__(self, settings_dict: typing.Optional[dict] = None):
    if settings_dict is None:
      settings_dict = settings.COMPETITION_SETTINGS
    self.settings = dict(DEFAULT_SETTINGS)
    self.settings.update(settings_dict)

  @classmethod
  def type_from_string(cls, service_type: str):
    """Load mail service from settings."""
    module, type_name = service_type.rsplit(".", 1)
    module = importlib.import_module(module)
    return getattr(module, type_name)

  @property
  def mail_service(self) -> typing.Type['AbstractMailService']:
    """Returns mail service."""
    return self.type_from_string(self.settings['MAIL_SERVICE'])

  @property
  def base_form(self) -> typing.Type:
    """Returns base form for voting."""
    return self.type_from_string(self.settings['COMPETITION_BASE_FORM'])

  @property
  def terms_url(self) -> str:
    """Returns terms and conditions url."""
    return self.settings['TERMS_AND_CONDITIONS_URL']

  @property
  def vote_registered_url(self):
    """Return vote redirect url after user """
    return self.settings['VOTE_REGISTERED_URL']

  @property
  def vote_confirm_success_url(self):
    """Return redirect after vote confirmation."""
    return self.settings['VOTE_SUCCESS_URL']

  @property
  def bleach_cleaner(self) -> bleach.Cleaner:
    """Returns cleaner used for sanitizing markdown"""
    return self.type_from_string(self.settings['BLEACH_CLEANER'])


class InvalidVoteException(Exception):
  """Error raised when we have invalid vote and error is an application"""
  pass


class UserInvalidVoteException(Exception):
  """Error raised when user made some disallowed action."""
  #                                     Not useless as we force message argument.
  def __init__(self, message) -> None:  # pylint: disable=useless-super-delegation
    super().__init__(message)


class CompetitionService(object):
  """Service handling competition."""

  def __init__(self, competition: 'models.Competition'):
    self.competition = competition

  def register_vote(self, vote: CompetitionVote):
    """Store vote in database."""
    assert self.competition == vote.competition
    with transaction.atomic():
      VoteService().validate_vote(vote)
      group = models.VoteGroup.objects.create(
        competition=vote.competition,
        used_nonce=vote.nonce,
        vote_source=vote.source_email
      )
      for entry in vote.entries:
        models.Vote.objects.create(group=group, entry=entry)

  @classmethod
  def recalculate_votes(cls, competition_id: int = None):
    """Recount all votes."""
    if isinstance(competition_id, models.Competition):
      competition_id = competition_id.pk

    queryset = models.CompetitionEntry.objects.all()
    if competition_id is not None:
      queryset = queryset.filter(competition__pk=competition_id)
    for entry in queryset.select_for_update():
      entry.votes = models.Vote.objects.filter(entry=entry).count()
      entry.save()


class VoteService(object):

  """Service for vote validation."""

  def validate_vote(self, vote: CompetitionVote):
    """Validates a vote."""
    self.__basic_validate_vote(vote)
    self.__validate_vote_with_competition(vote)

  def __validate_vote_with_competition(self, vote: CompetitionVote):
    """Helper. Validates a vote with competition requirements."""
    competition = vote.competition

    if not competition.voting_open:
      raise UserInvalidVoteException("Competition is closed.")

    if not MailChecker.is_valid_email_format(vote.source_email):
      raise UserInvalidVoteException("E-mail appears to be invalid")

    if not competition.allow_disposable_emails:
      if MailChecker.is_blacklisted(vote.source_email):
        raise UserInvalidVoteException(
          "You are using blacklisted e-mail provider"
        )

    if competition.entries_per_vote > 0:
      if competition.entries_per_vote < len(vote.entries):
        raise UserInvalidVoteException(ugettext_lazy(
          "You can vote for at most {} entries.".format(len(vote.entries))
        ))

  def __basic_validate_vote(self, vote: CompetitionVote):
    """
    Misc checks that verify vote was well-formed.
    """
    if not isinstance(vote.competition, models.Competition):
      raise InvalidVoteException()
    if not len(vote.entries) > 0:  # pylint: disable=len-as-condition
      raise InvalidVoteException()

    if not isinstance(vote.nonce, bytes):
      raise InvalidVoteException()

    all_entries = {entry.pk for entry in vote.competition.entries.all()}

    for entry in vote.entries:
      if not isinstance(entry, models.CompetitionEntry):
        raise InvalidVoteException()
      if entry.pk not in all_entries:
        raise InvalidVoteException()

    if len(vote.entries) != len({entry.pk for entry in vote.entries}):
      raise InvalidVoteException()

    conflicting_nonces = models.VoteGroup.objects.filter(
      competition=vote.competition,
      used_nonce=vote.nonce
    )
    if conflicting_nonces.exists():
      raise UserInvalidVoteException(
        "You have already confirmed this vote."
      )

  def serialize_vote(self, vote: CompetitionVote) -> bytes:
    """
    Serialize vote to opaque bytes (which can be signed).
    """
    vote_args = [
      ('competition', vote.competition.pk),
      ('nonce', vote.nonce),
      ('source_email', vote.source_email),
    ]
    vote_args.extend(sorted(("entry", str(entry.pk)) for entry in vote.entries))
    return urlencode(vote_args).encode('ascii')

  def deserialize_vote(self, vote: bytes) -> CompetitionVote:
    """De-serialize vote from bytes."""
    parsed = parse_qs(vote.decode('ascii'))

    return CompetitionVote(
      nonce=parsed['nonce'][0].encode('ascii'),
      source_email=parsed['source_email'][0],
      competition=models.Competition.objects.get(pk=int(parsed['competition'][0])),
      entries=list(models.CompetitionEntry.objects.filter(
        pk__in=[int(pk) for pk in parsed['entry']]
      ))
    )

  def sign_vote(
      self,
      competition: 'models.Competition',
      vote: bytes
  ) -> bytes:
    """Sign vote bytes."""
    hmac_obj = hmac.new(
      force_bytes(competition.competition_encryption_key),
      digestmod=hashlib.sha256
    )
    hmac_obj.update(vote)
    return hmac_obj.digest()

  def verify_vote_signature(
      self,
      competition: 'models.Competition',
      vote: bytes,
      signature: bytes
  ):
    """Verify vote signature."""
    expected_signature = self.sign_vote(competition, vote)
    if not hmac.compare_digest(signature, expected_signature):
      raise InvalidVoteException("Incorrect vote")

  def vote_to_signed_query_args(self, vote: CompetitionVote) -> typing.Dict:
    """Serialize vote to signed query arguments."""
    self.validate_vote(vote)
    serialized_vote = self.serialize_vote(vote)
    vote_signature = self.sign_vote(vote.competition, serialized_vote)
    return {
      "vote": base64.b64encode(serialized_vote),
      "sig": base64.b64encode(vote_signature)
    }

  def query_args_to_vote(self, query: typing.Dict) -> CompetitionVote:
    """De-serialize vote from signed query arguments."""
    try:
      vote_arg = query['vote']
      sig_arg = query['sig']
    except KeyError as key_error:
      raise InvalidVoteException from key_error

    vote_arg = vote_arg.encode('ascii') if isinstance(vote_arg, str) else vote_arg
    sig_arg = sig_arg.encode('ascii') if isinstance(sig_arg, str) else sig_arg

    if not all(isinstance(arg, bytes) for arg in[vote_arg, sig_arg]):
      raise InvalidVoteException()
    try:
      serialized_vote = base64.b64decode(vote_arg, validate=True)
      vote_signature = base64.b64decode(sig_arg, validate=True)
    except binascii.Error as decode_error:
      raise InvalidVoteException from decode_error

    vote = self.deserialize_vote(serialized_vote)
    self.verify_vote_signature(vote.competition, serialized_vote, vote_signature)

    self.validate_vote(vote)

    return vote


class AbstractMailService(object, metaclass=abc.ABCMeta):
  """Base class for a service that sends e-mails with confirmation."""
  def __init__(
      self,
      vote: CompetitionVote,
      confirm_url: str
  ):
    self.vote = vote
    self.confirm_url = confirm_url

  def send_confirm_mail(self):
    """Send confirmation e-mail to voter."""
    raise NotImplementedError


class SimpleMailService(AbstractMailService):
  """Base class for sending confirmation e-mails to users."""

  def get_subject(self) -> str:
    """Get message subject."""
    return ugettext_lazy("Vote confirmation for {}").format(self.vote.competition.name)

  def get_mail_render_context(self):
    """Get e-mail context"""
    return {
      "vote": self.vote,
      "vote_confirm_url": self.confirm_url
    }

  def get_mail_template(self):
    """Return mail template."""
    return get_template("django_competition/email.txt")

  def get_mail_contents(self):
    """Return rendered template."""
    return self.get_mail_template().render(context=self.get_mail_render_context())

  def get_email_from(self):
    """Sender address."""
    return settings.DEFAULT_FROM_EMAIL

  def send_confirm_mail(self):
    send_mail(
      subject=self.get_subject(),
      message=self.get_mail_contents(),
      from_email=self.get_email_from(),
      recipient_list=[self.vote.source_email]
    )


SETTINGS_SERVICE = SettingsService()


BLEACH_ALLOWED_TAGS = tuple(bleach.ALLOWED_TAGS + ['p'])
BLEACH_CLEANER = bleach.Cleaner(tags=BLEACH_ALLOWED_TAGS)



# coding=utf-8

"""Tests for vote service."""

import pytest

from django_competition import models, services
from django_competition.services import InvalidVoteException, UserInvalidVoteException


# pylint: disable=redefined-outer-name,invalid-name,no-member,unused-argument


@pytest.mark.django_db
def test_serialize_vote(
    service: services.VoteService,
    vote,
    serialized_vote
):
  """Checks vote serialized form (serialization part)."""
  assert service.serialize_vote(vote) == serialized_vote


@pytest.mark.django_db
def test_deserialize_vote(
    service: services.VoteService,
    vote,
    serialized_vote
):
  """Checks vote serialized form (deserialization part)."""
  assert service.deserialize_vote(serialized_vote) == vote


@pytest.mark.django_db
def test_vote_signature(
    service: services.VoteService,
    competition,
    serialized_vote,
    vote_signature
):
  """Test vote signature."""
  assert service.sign_vote(competition, serialized_vote) == vote_signature


@pytest.mark.django_db
def test_vote_verify_positive(
    service: services.VoteService,
    competition,
    serialized_vote,
    vote_signature
):
  """Test verification for valid vote."""
  service.verify_vote_signature(competition, serialized_vote, vote_signature)


@pytest.mark.django_db
def test_vote_verify_negative(
    service: services.VoteService,
    competition,
    serialized_vote,
    vote_signature
):
  """Test verification for invalid signature."""
  invalid_signature = b'\x01' * 32
  assert invalid_signature != vote_signature
  with pytest.raises(InvalidVoteException):
    service.verify_vote_signature(competition, serialized_vote, invalid_signature)


@pytest.mark.django_db
def test_vote_to_query(
    service: services.VoteService,
    vote
):
  """Test vote serialization and deserialization recovers the same vote."""
  query = service.vote_to_signed_query_args(vote)
  recovered_vote = service.query_args_to_vote(query)
  assert vote == recovered_vote


@pytest.mark.django_db
@pytest.mark.parametrize('query', [
  # Missing signature and vote
  {},
  {"foo": "bar"},
  # Invalid types
  {"sig": 1, "vote": object()},
  # Not base64 encoded signature and vote
  {"sig": b"foo", "vote": b"bar"}
])
def test_vote_to_query_invalid_args(
    service: services.VoteService,
    query
):
  """Test various malformed votes."""
  with pytest.raises(InvalidVoteException):
    service.query_args_to_vote(query)


@pytest.mark.django_db
def test_validate_invalid_email(service: services.VoteService, vote):
  """Validate vote that does not contain a valid e-mail."""
  with pytest.raises(UserInvalidVoteException):
    service.validate_vote(vote._replace(
      source_email="Not a valid e-mail"
    ))


@pytest.mark.django_db
def test_validate_disposable_email(service: services.VoteService, vote):
  """Test vote with disposable e-mail."""
  with pytest.raises(UserInvalidVoteException):
    service.validate_vote(vote._replace(
      source_email="test@niepodam.pl"
    ))


@pytest.mark.django_db
def test_validate_entries(service: services.VoteService, vote):
  """Test vote with too many entries."""
  with pytest.raises(UserInvalidVoteException):
    service.validate_vote(vote._replace(
      entries=list(models.CompetitionEntry.objects.all())
    ))


@pytest.mark.django_db
def test_validate_invalid_competition_type(service: services.VoteService, vote):
  """Test for invalid competition."""
  with pytest.raises(InvalidVoteException):
    service.validate_vote(vote._replace(competition=42))


@pytest.mark.django_db
def test_validate_invalid_entry_type(service: services.VoteService, vote):
  """Test for invalid entry."""
  with pytest.raises(InvalidVoteException):
    service.validate_vote(vote._replace(entries=[42]))


@pytest.mark.django_db
def test_validate_invalid_no_entries(service: services.VoteService, vote):
  """Test for no entries in vote."""
  with pytest.raises(InvalidVoteException):
    service.validate_vote(vote._replace(entries=[]))


@pytest.mark.django_db
def test_validate_invalid_entry_from_other_competition(
    service: services.VoteService,
    vote,
    other_entry
):
  """Test entry from other competition."""
  with pytest.raises(InvalidVoteException):
    service.validate_vote(vote._replace(entries=[other_entry]))


@pytest.mark.django_db
def test_validate_conflicting_nonce(service: services.VoteService, vote):
  """Test vote with conflicting nonce does not validate."""

  models.VoteGroup.objects.create(competition=vote.competition, used_nonce=vote.nonce)

  with pytest.raises(UserInvalidVoteException):
    service.validate_vote(vote)


@pytest.mark.django_db
def test_validate_duplicate_entries(service: services.VoteService, vote):
  """Test validate vote with duplicate entries."""
  with pytest.raises(InvalidVoteException):
    service.validate_vote(vote._replace(
      entries=vote.entries + vote.entries
    ))


@pytest.mark.django_db
@pytest.mark.parametrize('nonce', [1, 'foobarbaz', object()])
def test_vote_nonce_invalid(service: services.VoteService, vote, nonce):
  """Test vote with invalid nonce type."""
  with pytest.raises(InvalidVoteException):
    service.validate_vote(vote._replace(nonce=nonce))


@pytest.mark.django_db
def test_validate_competition_closed(
    service: services.VoteService,
    vote,
    competition,
):
  """Test validate case where vote is valid, but competition is closed."""
  competition.voting_open = False
  competition.save()

  with pytest.raises(UserInvalidVoteException):
    service.validate_vote(vote)




# coding=utf-8

"""Tests for competition form."""

import typing

import pytest
from django.forms import Form

from django_competition.forms import create_vote_form

# pylint: disable=redefined-outer-name


@pytest.fixture()
def competition_form(competition):
  """Type representing a form for this competition."""
  return create_vote_form(competition)


@pytest.mark.django_db
def test_choice_fields(competition_form):
  """Test choice form has fields for each vote."""
  form = competition_form()
  choice_fields = {
    field for field in form.fields.keys()
    if field.startswith("choice_")
  }
  assert choice_fields == {'choice_0', 'choice_1', 'choice_2'}


@pytest.mark.django_db
def test_valid_form(valid_competition_form_data, competition_form: typing.Type[Form]):
  """Tests a form with valid data validates."""
  form = competition_form(data=valid_competition_form_data)
  assert form.is_valid()


@pytest.mark.django_db
def test_invalid_form(invalid_competition_form_data, competition_form: typing.Type[Form]):
  """Multiple tests that check various inputs that could make competition form not validate."""
  data, broken_field = invalid_competition_form_data
  form = competition_form(data=data)
  assert not form.is_valid()
  assert form.errors.keys() == broken_field, form.errors




# coding=utf-8

from setuptools import find_packages, setup
import pathlib
import os
from pip.req import parse_requirements
from pip.download import PipSession

MAIN_DIR = pathlib.Path(__file__).absolute().parent


def get_requirements():
  pip_session = PipSession()
  return parse_requirements(
      str(MAIN_DIR / 'requirements' / 'base.in'),
      session=pip_session
  )

packages = find_packages(
  str(MAIN_DIR),
  include=('django_competition',),
  exclude=[]
)

if __name__ == "__main__":

  setup(
    name='django_competition',
    version='0.1.8',
    packages=packages,
    license='Apache',
    author='Jacek Bzdak',
    author_email='jacek@askesis.pl',
    description='Simple vote mechanism for a competition',
    install_requires=[
      "Django",
      "Markdown",
      "mailchecker",
      "bleach",
    ],
    include_package_data=True
  )



# coding=utf-8
"""Tests for views."""
import re
from urllib.parse import urlparse, parse_qs, urlencode

import pytest
from django.core import mail
from django.test import Client
from django.urls import reverse

from django_competition.models import VoteGroup
from django_competition.services import VoteService


# pylint: disable=redefined-outer-name,invalid-name,no-member,unused-argument


@pytest.fixture()
def vote_url(competition):
  """Returns url for page that allows voting on a competition."""
  return reverse("competition:vote", kwargs={"pk": competition.pk})


@pytest.fixture()
def valid_vote_response(
    valid_competition_form_data,
    vote_url,
    client,
):
  """HTTPS response to valid vote POST."""
  return client.post(vote_url, valid_competition_form_data)


@pytest.mark.django_db()
def test_vote_view_get(
    competition,
    client,
    vote_url,
):
  """Basic sanity checks in HTML response for voting view."""
  response = client.get(vote_url)
  body = response.content.decode('utf-8')
  assert competition.name in body
  for entry in competition.entries.all():
    assert entry.name in body


@pytest.mark.django_db()
def test_vote_view_post_single_vote(valid_vote_response):
  """Checks that post to vote pages is a redirect."""
  response = valid_vote_response
  assert response.status_code == 302
  assert response.url == reverse("competition:done")


@pytest.mark.django_db()
def test_after_vote_posted_email_is_sent(valid_vote_response):
  """Checks that post to vote pages results in an e-mail sent."""
  assert len(mail.outbox) == 1
  message_body = mail.outbox[0].body
  url_regex = re.compile(r"To confirm the vote go to the following page:\s*(.*)\s*\.\s*")
  match = re.findall(url_regex, message_body)
  assert len(match) == 1
  confirm_url = urlparse(match[0])
  query_args = parse_qs(confirm_url.query)

  service = VoteService()
  # This validates vote --- which means vote would be stored.
  service.query_args_to_vote({k: v[0] for k, v in query_args.items()})


@pytest.mark.django_db()
def test_invalid_vote_view(
    client: Client,
    vote_url,
    invalid_competition_form_data
):
  """Suite of tests for invalid votes."""
  data, __ = invalid_competition_form_data
  response = client.post(vote_url, data=data)
  # If form is invalid it is displayed again
  assert response.status_code == 200
  assert VoteGroup.objects.count() == 0


@pytest.fixture()
def valid_confirmation_url(
    vote,
):
  """Returns confirmation url for a vote."""
  service = VoteService()
  query_args = service.vote_to_signed_query_args(vote)
  return reverse("competition:confirm") + "?" + urlencode(query_args)


@pytest.mark.django_db()
def test_valid_vote_confirmation_get(
    client: Client,
    valid_confirmation_url,
    vote
):
  """Test for user entering a confirmation page. (No vote recorded)"""
  response = client.get(valid_confirmation_url)
  assert response.status_code == 200
  response_contents = response.content.decode('utf-8')
  assert "I confirm this vote" in response_contents
  assert vote.competition.name in response_contents
  for entry in vote.entries:
    assert entry.name in response_contents
  # Still no vote recorded
  assert VoteGroup.objects.count() == 0


@pytest.mark.django_db()
def test_valid_vote_confirmation_post(client: Client, valid_confirmation_url):
  """Tests POST on confirmation page confirms a vote."""
  response = client.post(path=valid_confirmation_url, data={})
  assert response.status_code == 302
  # Vote recorded
  assert VoteGroup.objects.count() == 1


@pytest.mark.django_db()
def test_invalid_vote_confirm(competition, client: Client, valid_confirmation_url):
  """Tests POST on confirmation page confirms a vote."""
  competition.voting_open = False
  competition.save()
  response = client.post(path=valid_confirmation_url, data={})
  assert response.status_code == 200
  assert "You can't confirm this vote right now" in response.content.decode('utf-8')
  assert VoteGroup.objects.count() == 0


@pytest.mark.django_db()
def test_malformed_vote_confirm(competition, client: Client):
  """Tests POST on confirmation page confirms a vote."""
  competition.voting_open = False
  competition.save()
  #                           No vote parameters so invalid vote
  response = client.post(path=reverse("competition:confirm"), data={})
  assert response.status_code == 400





"""Admin module."""

from django.contrib import admin

from . import models


class CompetitionEntry(admin.TabularInline):
  """Competition Entry in-line admin."""
  model = models.CompetitionEntry

  fields = [
    'name',
    'description',
    'votes',
  ]

  readonly_fields = [
    'votes'
  ]


@admin.register(models.Competition)
class CompetitionAdmin(admin.ModelAdmin):
  """Competition model admin."""

  list_display = [
    'name'
  ]

  fields = [
    'name',
    'description',
    'votes_per_day',
    'entries_per_vote',
    'allow_disposable_emails'
  ]

  inlines = [
    CompetitionEntry
  ]


@admin.register(models.Vote)
class VoteAdmin(admin.ModelAdmin):

  fields = readonly_fields = [
    'entry', 'group'
  ]

  list_display = [
    'entry', 'vote_competition', 'vote_email'
  ]

  list_filter = [
    'group__competition'
  ]

  def vote_competition(self, obj):
    return obj.group.competition

  def vote_email(self, obj):
    return obj.group.vote_source



"""Models."""

import os

from django.core.validators import MinValueValidator
from django.db import models, transaction
from django.db.models.signals import post_save, post_delete
from django.dispatch import receiver
from django.utils.safestring import SafeString
from django.utils.translation import ugettext_lazy
from markdown import markdown

from django_competition.services import SETTINGS_SERVICE


def random_key() -> bytes:
  """Generate random key for competition."""
  return os.urandom(32)


class Competition(models.Model):
  """Configuration for a single competition."""

  name = models.CharField(max_length=128)
  description = models.TextField()

  votes_per_day = models.IntegerField(
    default=-1,
    help_text=ugettext_lazy(
      "If > 0 allows single e-mail to vote this many votes in each day."
      "If <= 0 allows only single vote."
    )
  )

  entries_per_vote = models.PositiveIntegerField(
    default=3,
    blank=False,
    validators=[MinValueValidator(1)],
    help_text=ugettext_lazy("How many (equally weighted) entries you may select in each vote. ")
  )

  allow_disposable_emails = models.BooleanField(
    default=True,
    blank=True,
    help_text=ugettext_lazy("If false we will try to block disposable e-mails.")
  )

  competition_encryption_key = models.BinaryField(
    default=random_key,
    editable=False,
    help_text=ugettext_lazy("Encryption key used to sign vote requests.")
  )

  voting_open = models.BooleanField(
    verbose_name=ugettext_lazy("True if users can vote on the competition"),
    default=True,
  )

  @property
  def description_html(self):
    """Return HTML from rendered markdown."""
    return SafeString(SETTINGS_SERVICE.bleach_cleaner.clean(markdown(self.description)))

  def __str__(self):
    return self.name


class CompetitionEntry(models.Model):

  """
  Competition entry.

  This can be voted on.
  """

  competition = models.ForeignKey(
    Competition,
    related_name="entries",
    on_delete=models.CASCADE
  )
  name = models.CharField(max_length=256)
  description = models.TextField(
    help_text=ugettext_lazy("You can use some markdown here.")
  )
  votes = models.IntegerField(default=0)
  """
  De-normalized vote count. This can be derived by counting all Votes attached
  to this entry.
  """

  @property
  def description_html(self):
    """Markdown rendered HTML."""
    return SafeString(SETTINGS_SERVICE.bleach_cleaner.clean(markdown(self.description)))

  def __str__(self):
    return self.name


class VoteGroup(models.Model):
  """
  Vote group.

  Represents a singe vote from a single user, but this vote might be for
  more than one entries.

  Each vote for each entry is stored in Vote model.
  """

  class Meta:
    ordering = ['competition', 'date_voted']

  competition = models.ForeignKey(Competition)
  vote_source = models.TextField(null=False, blank=False)
  """
  Identifier of voting entity. Most often e-mail ;)
  """
  date_voted = models.DateTimeField(null=False, blank=False, auto_now_add=True)

  used_nonce = models.BinaryField(
    db_index=True
  )


class Vote(models.Model):
  """
  Single vote for a single competition entry.
  """

  class Meta:
    ordering = ['group__competition', 'entry', 'group__date_voted']

  group = models.ForeignKey(
    VoteGroup,
    related_name="votes",
    on_delete=models.CASCADE
  )

  entry = models.ForeignKey(
    CompetitionEntry,
    related_name="entries",
    on_delete=models.CASCADE
  )


@receiver(post_save, sender=Vote)
def on_vote_save(instance: Vote, **kwargs):  # pylint: disable=unused-argument
  """On vote save add vote to entry."""
  with transaction.atomic():
    entry = CompetitionEntry.objects.select_for_update().get(pk=instance.entry_id)
    entry.votes += 1
    entry.save(update_fields=['votes'])


@receiver(post_delete, sender=Vote)
def on_vote_delete(instance: Vote, **kwargs):  # pylint: disable=unused-argument
  """On vote delete remove vote from entry."""
  with transaction.atomic():
    entry = CompetitionEntry.objects.select_for_update().get(pk=instance.entry_id)
    entry.votes -= 1
    entry.save(update_fields=['votes'])
